import configparser
import logging

from neo4j import GraphDatabase, Driver
from pandas import DataFrame
from run import args

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Neo4jConnection:
    # class-level attribute
    __instance = None
    __uri = None
    __username = None
    __password = None

    def __new__(cls, *arg, **kwargs):
        # definition as singleton
        if not cls.__instance:
            cls.__instance = super(Neo4jConnection, cls).__new__(cls, *arg, **kwargs)
        return cls.__instance

    def __init__(self) -> None:
        # Connect to Neo4j database
        if not hasattr(self, '__initialized'):
            if not self.__uri:
                # read config file
                conf_file = configparser.ConfigParser()
                conf_file.read(args.conf)

                self.__uri = conf_file['neo4j']['uri']
                self.__username = conf_file['neo4j']['username']
                self.__password = conf_file['neo4j']['password']

            # instance-level attribute for the __connection, will be initialised later
            self.__connection = None
            # Establish __connection when the instance is created
            self.__start_connection()
            self.__initialized = True

    def __start_connection(self):
        """
        Creating a Driver instance only provides information on how to access the database, but does not actually
        establish a __connection. Connection is instead deferred to when the first query is executed.
        """
        try:
            driver: Driver = GraphDatabase.driver(self.__uri,
                                                  auth=(self.__username, self.__password))  # creation of Driver object
            driver.verify_connectivity()
            self.__connection = driver
            logger.debug("Connected to Neo4j database")
            return self.__connection
        except Exception as e:
            logger.error(f"Failed to connect to Neo4j database: {e}")
            raise

    def end_connection(self) -> None:
        if self.__connection is not None:
            self.__connection.close()
            logger.debug("Connection to Neo4j database closed")
            self.__connection = None

    def create_study_node(self, data: DataFrame, modelid):
        """
        based on https://neo4j.com/docs/python-manual/current/query-simple/
        """
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MERGE (st:Study {OID: $studyoid, Name: $studyname, Description: $studydescr, Protocol: $studyprotocol, ModelID: $modelid})",
                    studyoid=row['oid'], studyname=row['name'], studydescr=row['descr'], studyprotocol=row["protocol"], modelid=modelid
                )
                logger.debug(f"Inserted Study node with OID {row['oid']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_studyevent_node(self, data: DataFrame, studyoid: str, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query("MERGE (:StudyEvent {OID: $oid, Name: $name, Type: $type, StudyOID: $study_oid, ModelID: $modelid})",
                                   oid=row['OID'], name=row['Name'], type=row['Type'], study_oid=studyoid, modelid=modelid
                                   )
                logger.debug(f"Inserted StudyEvent node with OID {row['OID']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_form_node(self, data: DataFrame, studyoid: str, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MERGE (:Form {OID: $oid, Name: $name, StudyOID: $studyoid, ModelID: $modelid})",
                    oid=row['OID'], name=row['Name'], studyoid=studyoid, modelid=modelid)
                logger.debug(f"Inserted Form node with OID {row['OID']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_itemgroup_node(self, data: DataFrame, studyoid: str, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MERGE (:ItemGroup {OID: $oid, Name: $name, StudyOID: $studyoid, ModelID: $modelid})",
                    oid=row['OID'], name=row['Name'], studyoid=studyoid, modelid=modelid)
                logger.debug(f"Inserted ItemGroup node with OID {row['OID']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_item_node(self, data: DataFrame, studyoid: str, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MERGE (:Item {OID: $oid, Name: $name, DataType: $datatype, Question: $question, StudyOID: $studyoid, ModelID: $modelid})",
                    oid=row['itemoid'], name=row['name'], datatype=row['datatype'], question=row['question'],
                    studyoid=studyoid, modelid=modelid)
                logger.debug(f"Inserted Item node with OID {row['itemoid']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_alias_node(self, data: DataFrame,
                          studyoid: str, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MERGE (:Alias {Context: $context, Name: $name, ItemOID: $item, StudyOID: $study, ModelID: $modelid})",
                    context=row['end'], name=row['end_two'], item=row['start'], study=studyoid, modelid=modelid)
                logger.debug(f"Inserted Alias node with name {row['end_two']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_measurementunit_node(self, data: DataFrame):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MERGE (:MeasurementUnit {OID: $oid, Name: $name, Symbol: $symbol})",
                    oid=row['oid'], name=row['name'], symbol=row['symbol'])
                logger.debug(f"Inserted MeasurementUnit node with name {row['name']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_basedef_node(self, studyoid: str, modelid):
        conn = self.__connection
        try:
            conn.execute_query(
                "MERGE (:BasicDefinitions {StudyOID: $studyoid, ModelID: $modelid})", studyoid=studyoid, modelid=modelid)
            logger.debug(f"Inserted BasicDefinitions node for study {studyoid}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_rangecheck_node(self, data: DataFrame):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MERGE (:RangeCheck {Comparator: $comp, Constraint: $cons, CheckValue: $val})",
                    comp=row['comparator'], cons=row['constraint'], val=row['value'])
                logger.debug(f"Inserted RangeCheck node")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_codelist_node(self, data: DataFrame, studyoid, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MERGE (:CodeList {OID: $oid, Name: $name, DataType: $type, StudyOID: $studyoid, ModelID: $modelid})",
                    oid=row['OID'], name=row['Name'], type=row['DataType'], studyoid=studyoid, modelid=modelid)
                logger.debug(f"Inserted CodeList node")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_clitem_node(self, data: DataFrame):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MERGE (:CodeListItem {CodedValue: $cv, Decode: $dc})",
                    cv=row['codedvalue'], dc=row['decode'])
                logger.debug(f"Inserted CodeListItem node")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_study_to_studyevent_rel(self, studyoid: str, modelid):
        conn = self.__connection
        try:
            conn.execute_query(
                "MATCH (start:Study {OID: $studyoid, ModelID: $modelid}), (end:StudyEvent {StudyOID: $studyoid, ModelID: $modelid}) MERGE (start)-[:STUDY_HAS_STUDYEVENT]->(end)-[:BACKTRACK]->(start)",
                studyoid=studyoid, modelid=modelid
            )
            logger.debug(f"Inserted rels betw Study and StudyEvent")
        except Exception as e:
            logger.error(f"Failed to populate Neo4j database: {e}")
            raise

    def create_itemgroup_to_item_rel(self, data: DataFrame, studyoid: str, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MATCH (start:ItemGroup {OID: $startoid, StudyOID: $studyoid, ModelID: $modelid}), (end:Item {OID: $endoid, StudyOID: $studyoid, ModelID: $modelid}) MERGE (start)-[:ITEMGROUP_HAS_ITEM]->(end)-[:BACKTRACK]->(start)",
                    startoid=row['start'], endoid=row['end'], studyoid=studyoid, modelid=modelid)
                logger.debug(
                    f"Inserted rels betw ItemGroup {row['start']} and Item {row['end']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_form_to_itemgroup_rel(self, data: DataFrame, studyoid: str, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MATCH (start:Form {OID: $startoid, StudyOID: $studyoid, ModelID: $modelid}), (end:ItemGroup {OID: $endoid, StudyOID: $studyoid, ModelID: $modelid}) MERGE (start)-[:FORM_HAS_ITEMGROUP]->(end)-[:BACKTRACK]->(start)",
                    startoid=row['start'], endoid=row['end'], studyoid=studyoid, modelid=modelid)
                logger.debug(
                    f"Inserted rels betw Form {row['start']} and ItemGroup {row['end']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_studyevent_to_form_rel(self, data: DataFrame, studyoid: str, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MATCH (start:StudyEvent {OID: $startoid, StudyOID: $studyoid, ModelID: $modelid}), (end:Form {OID: $endoid, StudyOID: $studyoid, ModelID: $modelid}) MERGE (start)-[:STUDYEVENT_HAS_FORM]->(end)-[:BACKTRACK]->(start)",
                    startoid=row['start'], endoid=row['end'], studyoid=studyoid, modelid=modelid)
                logger.debug(
                    f"Inserted rels betw StudyEvent {row['start']} and Form {row['end']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_item_to_alias_rel(self, data: DataFrame, studyoid: str, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MATCH (start:Item {OID: $startoid, StudyOID: $studyoid, ModelID: $modelid}), (end:Alias {Context: $endcontext, Name: $endname, ItemOID: $startoid, StudyOID: $studyoid, ModelID: $modelid}) MERGE (start)-[:ITEM_HAS_ALIAS]->(end)-[:BACKTRACK]->(start)",
                    startoid=row['start'], endcontext=row['end'], endname=row['end_two'], studyoid=studyoid, modelid=modelid)
                logger.debug(
                    f"Inserted rels betw Item {row['start']} and Alias {row['end']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_itemgroup_to_alias_rel(self, data: DataFrame, studyoid: str, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MATCH (start:ItemGroup {OID: $startoid, StudyOID: $studyoid, ModelID: $modelid}), (end:Alias {Context: $endcontext, Name: $endname}) MERGE (start)-[:ITEMGROUP_HAS_ALIAS]->(end)-[:BACKTRACK]->(start)",
                    startoid=row['start'], endcontext=row['end'], endname=row['end_two'], studyoid=studyoid, modelid=modelid)
                logger.debug(
                    f"Inserted rels betw ItemGroup {row['start']} and Alias {row['end']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_item_to_measurement_rel(self, data: DataFrame, studyoid: str, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MATCH (start:Item {OID: $startoid, StudyOID: $studyoid, ModelID: $modelid}), (end:MeasurementUnit {OID: $endoid, Name: $endname}) MERGE (start)-[:ITEM_HAS_MEASUREMENTUNIT]->(end)-[:BACKTRACK]->(start)",
                    startoid=row['start'], endoid=row['endoid'], endname=row['endname'], studyoid=studyoid, modelid=modelid)
                logger.debug(
                    f"Inserted rels betw Item {row['start']} and MeasurementUnit {row['endname']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_basedef_to_study_rel(self, studyoid: str, modelid):
        conn = self.__connection
        try:
            conn.execute_query(
                "MATCH (start:Study {OID: $studyoid, ModelID: $modelid}), (end:BasicDefinitions {StudyOID: $studyoid, ModelID: $modelid}) MERGE (start)-[:HAS_BASEDEF]->(end)",
                studyoid=studyoid, modelid=modelid)
            logger.debug(
                f"Inserted rels betw Study {studyoid} and BasicDefinitions")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_basedef_to_measurement_rel(self, data: DataFrame,
                                          studyoid: str, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MATCH (start:BasicDefinitions {StudyOID: $studyoid, ModelID: $modelid}), (end:MeasurementUnit {OID: $endoid, Name: $endname, Symbol: $endsymbol}) MERGE (start)-[:BASEDEF_HAS_MEASUREMENTUNIT]->(end)",
                    endoid=row['oid'], endname=row['name'], endsymbol=row['symbol'], studyoid=studyoid, modelid=modelid)
                logger.debug(
                    f"Inserted rels betw BasicDefinitions and MeasurementUnit {row['name']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_item_to_rangecheck_rel(self, data: DataFrame):  #TODO check for Item properties (missing StudyOID?)
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MATCH (start:Item {OID: $itemoid}), (end:RangeCheck {Comparator: $comp, Constraint: $cons, CheckValue: $cvalue}) MERGE (start)-[:ITEM_HAS_RANGECHECK]->(end)-[:BACKTRACK]->(start)",
                    itemoid=row['start'], comp=row['endcomparator'], cons=row['endconstraint'], cvalue=row['endvalue'])
                logger.debug(
                    f"Inserted rels betw Item and RangeCheck")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_item_to_codelist_rel(self, data: DataFrame, studyoid, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MATCH (start:Item {OID: $startoid, StudyOID: $studyoid, ModelID: $modelid}), (end:CodeList {OID: $endoid, StudyOID: $studyoid, ModelID: $modelid}) MERGE (start)-[:ITEM_HAS_CODELIST]->(end)-[:BACKTRACK]->(start)",
                    startoid=row['start'], endoid=row['end'], studyoid=studyoid, modelid=modelid)
                logger.debug(
                    f"Inserted rels betw Item {row['start']} and CodeList {row['end']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def create_codelist_to_codelistitem_rel(self, data: DataFrame, studyoid, modelid):
        conn = self.__connection
        try:
            for index, row in data.iterrows():
                conn.execute_query(
                    "MATCH (start:CodeList {OID: $startoid, StudyOID: $studyoid, ModelID: $modelid}), (end:CodeListItem {CodedValue: $cv, Decode: $dc}) MERGE (start)-[:CODELIST_HAS_CODELISTITEM]->(end)-[:BACKTRACK]->(start)",
                    startoid=row['start'], cv=row['endcodedvalue'], dc=row['enddecode'], studyoid=studyoid, modelid=modelid)
                logger.debug(
                    f"Inserted rels betw CodeList {row['start']} and CodeListItem {row['endcodedvalue']}")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    # postprocessing with cypher queries
    def postprocess_set_label_post(self):
        conn = self.__connection
        try:
            conn.execute_query(
                "MATCH (n:Alias) WHERE n.Context contains '[' and n.Context contains ','  and n.Context contains ']' set n :Post"
            )
            logger.debug(f"Inserted post property in alias nodes")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise

    def postprocess_connect_postcoord_aliases(self, tuples):
        conn = self.__connection
        try:
            conn.execute_query(
                """
                UNWIND $data AS row
                MATCH (i:Item {OID: row.item_oid})
                WITH row, i
                MATCH (a1:Post {StudyOID: row.study_oid, Context: row.ontology + '[' + row.x + ',' + 1 + ']'})--(i)
                MATCH (a2:Post {StudyOID: row.study_oid, Context: row.ontology + '[' + row.x + ',' + row.y + ']'})--(i)
                WHERE a1 <> a2
                MERGE (a1)-[rel:COMPOSITE]->(a2)
            """, data=tuples)
            logger.debug(f"Inserted rels btw post-coordinated aliases")
        except Exception as e:
            logger.error(f"Failed to execute query: {e}")
            raise
