# this script extracts data of an xml file and stores the data in a pandas DataFrame
# the DataFrame is to be given to the node and relationship creator methods from database_connector.py

import logging
import os

import pandas as pd
from bs4 import BeautifulSoup
from pandas import DataFrame
import re

import database_connector.database_connector as db

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class XmlProcessor:
    """
    class handles one XML file
    """

    def __init__(self, xml: BeautifulSoup):
        self.xmlfile_to_process = xml

    # helper functions to create multiple data stores for nodes
    # general function to create data stores for nodes
    @staticmethod
    def __extract_data_to_store_for_nodes(soup: object,
                                          tag_name: str,
                                          attributes: list):
        """
        :param soup: parsed xml file (parsed with beautiful soup, somewhere)
        :param tag_name: defines for which node label the data is stored, e.g., Alias
        :param attributes: one of the lists from above, depending on tag_name, e.g., alias_attributes for Alias node
        :return: pandas Dataframe with data extracted from xml that is needed for creation of nodes in neo4j
        """
        elements = soup.findAll(tag_name)
        data = {attr: [] for attr in attributes}
        for element in elements:
            for attr in attributes:
                data[attr].append(element.get(attr))
        element_data = pd.DataFrame(data)
        return element_data

    # funtion to create data store for study nodes
    @staticmethod
    def __get_study_data(soup):
        """
        :param soup: parsed xml file (parsed with beautiful soup, somewhere)
        :return: pandas Dataframe with data extracted from xml that is needed for creation of nodes with label Study in neo4j
        """
        # global studyprotocol_list, studydescr_list, studyname_list
        study = soup.find('Study')  # find 1 tag
        studyoid = study.get('OID')
        if studyoid == "some OID":
            odm = soup.find('ODM')
            studyoid_list = [odm.get('Description')]
        else:
            studyoid_list = [study.get('OID')]
        # studyoid_list = [study.get('OID')]
        for child in study.GlobalVariables:
            if child.name == 'StudyName':
                studyname_list = [child.get_text()]
            elif child.name == 'StudyDescription':
                studydescr_list = [child.get_text()]
            elif child.name == 'ProtocolName':
                studyprotocol_list = [child.get_text()]

        # studydata = []
        studydata = pd.DataFrame({"oid": studyoid_list,
                                  "name": studyname_list,
                                  "descr": studydescr_list,
                                  "protocol": studyprotocol_list})
        return studydata

    # funtion to create data store for itemdef nodes with question property
    @staticmethod
    def __itemdef_data_store_with_question(soup):
        """
        :param soup: parsed xml file (parsed with beautiful soup, somewhere)
        :return: pandas Dataframe with data extracted from xml that is needed for creation of nodes with label ItemDef
        """
        itemdef = soup.findAll('ItemDef')
        itemdef_itemoid = []
        itemdef_names = []
        itemdef_datatypes = []
        itemdef_questions = []
        for item in itemdef:
            itemdef_itemoid.append(item.get('OID'))
            itemdef_names.append(item.get('Name'))
            itemdef_datatypes.append(item.get('DataType'))

            question = item.find('Question')

            translated_text = question.find('TranslatedText', {'xml:lang': 'de'})

            if not translated_text:
                translated_text = question.find('TranslatedText', {'xml:lang': 'en'})

            itemdef_questions.append(translated_text.text)

        data = pd.DataFrame({"itemoid": itemdef_itemoid,
                             "name": itemdef_names,
                             "datatype": itemdef_datatypes,
                             "question": itemdef_questions})
        return data

    @staticmethod
    def __measurementunit_data_store(soup):
        """
        :param soup: parsed xml file (parsed with beautiful soup, somewhere)
        :return: pandas Dataframe with data extracted from xml that is needed for creation of nodes with label ItemDef
        """
        measurementunit = soup.findAll('MeasurementUnit')
        measurementunit_oid = []
        measurementunit_names = []
        measurementunit_symbols = []
        for item in measurementunit:
            measurementunit_oid.append(item.get('OID'))
            measurementunit_names.append(item.get('Name'))

            symbol = item.find('Symbol')

            translated_text = symbol.find('TranslatedText', {'xml:lang': 'de'})

            if not translated_text:
                translated_text = symbol.find('TranslatedText', {'xml:lang': 'en'})

            measurementunit_symbols.append(translated_text.text)

        data = pd.DataFrame({"oid": measurementunit_oid,
                             "name": measurementunit_names,
                             "symbol": measurementunit_symbols})
        return data

    @staticmethod
    def __get_rangecheck_data(soup):
        """
        :param soup: parsed xml file (parsed with beautiful soup, somewhere)
        :return: pandas Dataframe with data extracted from xml that is needed for creation of nodes with label RangeCheck in neo4j
        """
        rangecheck = soup.findAll('RangeCheck')
        rc_comparators = []
        rc_constraints = []
        rc_values = []
        for rc in rangecheck:
            rc_comparators.append(
                rc.get('Comparator', 'no entry'))
            rc_constraints.append(rc.get('SoftHard'))

            # Collect all CheckValue elements
            checkvalues = [child.get_text() for child in rc.find_all('CheckValue')]
            # If no CheckValue elements exist, use default
            if not checkvalues:
                checkvalues = ['empty list']
            rc_values.append(checkvalues)

        rc_data = pd.DataFrame({"comparator": rc_comparators,
                                "constraint": rc_constraints,
                                "value": rc_values})
        return rc_data

    @staticmethod
    def __codelistitem_data_store(soup):
        """
        :param soup: parsed xml file (parsed with beautiful soup, somewhere)
        :return: pandas Dataframe with data extracted from xml that is needed for creation of nodes with label ItemDef
        """
        clitem = soup.findAll('CodeListItem')
        clitem_codedvalues = []
        clitem_decodes = []
        for cl in clitem:
            clitem_codedvalues.append(cl.get('CodedValue'))
            decode = cl.find('Decode')

            translated_text = decode.find('TranslatedText', {'xml:lang': 'de'})

            if not translated_text:
                translated_text = decode.find('TranslatedText', {'xml:lang': 'en'})

            clitem_decodes.append(translated_text.text)

        data = pd.DataFrame({"codedvalue": clitem_codedvalues,
                             "decode": clitem_decodes})
        return data

    # general function to create data stores for rels
    @staticmethod
    def __extract_data_to_store_for_relationships(soup,
                                                  def_tag_name: str,
                                                  ref_tag_name: str,
                                                  def_attr_name: str,
                                                  ref_attr_name: str):
        """
        :param soup: parsed xml file (parsed with beautiful soup, somewhere)
        :param def_tag_name: startnode label, e.g., label "ItemGroup", input is then "ItemGroupDef" (because of xml structure)
        :param ref_tag_name: endnode label, e.g., label "Item", input is then "ItemDef"
        :param def_attr_name: startnode property for finding right endnodes (e.g., find all Items that belong to ItemGroup)
        :param ref_attr_name: endnode property that is referenced in startnode (e.g., Items referenced in ItemGroup)
        :return: pandas Dataframe with data extracted from xml that is needed for creation of relationships in neo4j
        """
        end_elements = soup.findAll(ref_tag_name)
        data = {'start': [], 'end': []}
        for end_element in end_elements:
            parents = end_element.parents
            for p in parents:
                if p.name == def_tag_name:
                    data['end'].append(end_element.get(ref_attr_name))
                    data['start'].append(p.get(def_attr_name))
        rels_data = pd.DataFrame(data)
        return rels_data

    @staticmethod
    def __extract_data_with_two_attributes_to_store_for_relationships(soup,
                                                                      def_tag_name: str,
                                                                      ref_tag_name: str,
                                                                      def_attr_name: str,
                                                                      ref_attr_name: str,
                                                                      ref_attr_name_2: str):
        """
        :param soup: parsed xml file (parsed with beautiful soup, somewhere)
        :param def_tag_name: startnode label, e.g., label "ItemGroup", input is then "ItemGroupDef" (because of xml structure)
        :param ref_tag_name: endnode label, e.g., label "Item", input is then "ItemDef"
        :param def_attr_name: startnode property for finding right endnodes (e.g., find all Items that belong to ItemGroup)
        :param ref_attr_name: endnode property that is referenced in startnode (e.g., Items referenced in ItemGroup)
        :return: pandas Dataframe with data extracted from xml that is needed for creation of relationships in neo4j
        """
        end_elements = soup.findAll(ref_tag_name)
        data = {'start': [], 'end': [], 'end_two': []}
        for end_element in end_elements:
            parents = end_element.parents
            for p in parents:
                if p.name == def_tag_name:
                    data['end'].append(end_element.get(ref_attr_name))
                    data['end_two'].append(end_element.get(ref_attr_name_2))
                    data['start'].append(p.get(def_attr_name))
        rels_data = pd.DataFrame(data)
        return rels_data

    @staticmethod
    def __extract_data_for_item_to_alias_rels(soup):
        end_elements = soup.findAll('Alias')
        data = {'start': [], 'endcontext': [], 'endname': []}
        for element in end_elements:
            parents = end_elements.parents
            for p in parents:
                if p.name == 'ItemDef':
                    data['endcontext'].append(element.get('Context'))
                    data['endname'].append(element.get('Name'))
                    data['start'].append(p.get('OID'))
        rels_data = pd.DataFrame(data)
        return rels_data

    @staticmethod
    def __extract_data_for_to_alias_rels(soup, def_tag_name: str):
        end_elements = soup.findAll('Alias')
        data = {'start': [], 'endcontext': [], 'endname': []}
        for element in end_elements:
            parents = end_elements.parents
            for p in parents:
                if p.name == def_tag_name:
                    data['endcontext'].append(element.get('Context'))
                    data['endname'].append(element.get('Name'))
                    data['start'].append(p.get('OID'))
        rels_data = pd.DataFrame(data)
        return rels_data

    @staticmethod
    def __extract_data_for_item_to_measurement_rels(soup):
        end_elements = soup.findAll('MeasurementUnitRef')
        data = {'start': [], 'endoid': [], 'endname': []}
        for end_element in end_elements:
            parents = end_element.parents
            for p in parents:
                if p.name == 'ItemDef':
                    moid = end_element.get('MeasurementUnitOID')
                    endnameplaceholder = soup.find('MeasurementUnit', {'OID': moid})
                    mname = endnameplaceholder.get('Name')
                    data['endname'].append(mname)
                    data['endoid'].append(moid)
                    data['start'].append(p.get('OID'))
        rels_data = pd.DataFrame(data)
        return rels_data

    @staticmethod
    def __extract_data_for_item_to_checkvalue_rels(soup):
        end_elements = soup.findAll('RangeCheck')
        data = {'start': [], 'endcomparator': [], 'endconstraint': [], 'endvalue': []}
        for end_element in end_elements:
            parents = end_element.parents
            for p in parents:
                if p.name == 'ItemDef':
                    comparator = end_element.get('Comparator', 'null')
                    soft_hard = end_element.get('SoftHard', 'null')
                    check_value_elements = [cv.get_text() for cv in end_element.find_all('CheckValue')]
                    check_value = check_value_elements if check_value_elements else ['null']

                    data['endcomparator'].append(comparator)
                    data['endconstraint'].append(soft_hard)
                    data['endvalue'].append(check_value)

                    data['start'].append(p.get('OID'))
        rels_data = pd.DataFrame(data)
        return rels_data

    @staticmethod
    def __extract_data_for_codelist_to_codelistitem_rels(soup):
        end_elements = soup.findAll('CodeListItem')
        data = {'start': [], 'endcodedvalue': [], 'enddecode': []}
        for end_element in end_elements:
            parents = end_element.parents
            for p in parents:
                if p.name == 'CodeList':
                    codedvalue = end_element.get('CodedValue', 'null')

                    decode = end_element.find('Decode')
                    translated_text = decode.find('TranslatedText', {'xml:lang': 'de'})

                    if not translated_text:
                        translated_text = decode.find('TranslatedText', {'xml:lang': 'en'})

                    data['endcodedvalue'].append(codedvalue)
                    data['enddecode'].append(translated_text.text)

                    data['start'].append(p.get('OID'))
        rels_data = pd.DataFrame(data)
        return rels_data

    # process xml file with the help of the helper functions
    def get_studyoid(self):
        return XmlProcessor.__get_study_data(self.xmlfile_to_process)['oid'][0]

    def extract_study_data(self) -> DataFrame:
        study_data: DataFrame = XmlProcessor.__get_study_data(self.xmlfile_to_process)
        return study_data

    def extract_item_data(self) -> DataFrame:
        itemdef_data: DataFrame = XmlProcessor.__itemdef_data_store_with_question(self.xmlfile_to_process)
        return itemdef_data

    def extract_studyevent_data(self) -> DataFrame:
        studyeventdef_attributes = ['OID', 'Name', 'Type']
        studyeventdef_data: DataFrame = XmlProcessor.__extract_data_to_store_for_nodes(self.xmlfile_to_process,
                                                                                       'StudyEventDef',
                                                                                       studyeventdef_attributes)
        return studyeventdef_data

    def extract_form_data(self) -> DataFrame:
        formdef_attributes = ['OID', 'Name']
        formdef_data: DataFrame = XmlProcessor.__extract_data_to_store_for_nodes(self.xmlfile_to_process,
                                                                                 'FormDef',
                                                                                 formdef_attributes)
        return formdef_data

    def extract_itemgroup_data(self) -> DataFrame:
        itemgroupdef_attributes = ['OID', 'Name']
        itemgroupdef_data: DataFrame = XmlProcessor.__extract_data_to_store_for_nodes(self.xmlfile_to_process,
                                                                                      'ItemGroupDef',
                                                                                      itemgroupdef_attributes)
        return itemgroupdef_data

    def extract_alias_data(self):
        alias_attributes = ['Context', 'Name']
        alias_data: DataFrame = XmlProcessor.__extract_data_to_store_for_nodes(self.xmlfile_to_process, 'Alias',
                                                                               alias_attributes)
        return alias_data

    def extract_measurement_data(self):
        measurement_data: DataFrame = XmlProcessor.__measurementunit_data_store(self.xmlfile_to_process)
        return measurement_data

    def extract_rangecheck_data(self):
        rangecheck_data: DataFrame = XmlProcessor.__get_rangecheck_data(self.xmlfile_to_process)
        return rangecheck_data

    def extract_codelist_data(self):
        codelist_attributes = ['OID', 'Name', 'DataType']
        codelist_data: DataFrame = XmlProcessor.__extract_data_to_store_for_nodes(self.xmlfile_to_process, 'CodeList',
                                                                                  codelist_attributes)
        return codelist_data

    def extract_codelistitem_data(self):
        clitem_data: DataFrame = XmlProcessor.__codelistitem_data_store(self.xmlfile_to_process)
        return clitem_data

    @property
    def extract_itemgroup_to_item_data(self):
        itemgroup_to_item_data: DataFrame = XmlProcessor.__extract_data_to_store_for_relationships(
            self.xmlfile_to_process,
            'ItemGroupDef', 'ItemRef',
            'OID', 'ItemOID')
        return itemgroup_to_item_data

    def extract_item_to_alias_data(self) -> DataFrame:
        item_to_alias_data: DataFrame = XmlProcessor.__extract_data_with_two_attributes_to_store_for_relationships(
            self.xmlfile_to_process,
            'ItemDef',
            'Alias',
            'OID',
            'Context',
            'Name')
        return item_to_alias_data

    def extract_itemgroup_to_alias_data(self) -> DataFrame:
        itemgroup_to_alias_data: DataFrame = XmlProcessor.__extract_data_with_two_attributes_to_store_for_relationships(
            self.xmlfile_to_process,
            'ItemGroupDef',
            'Alias',
            'OID',
            'Context',
            'Name')
        return itemgroup_to_alias_data

    def extract_form_to_itemgroup_data(self) -> DataFrame:
        form_to_itemgroup_data: DataFrame = XmlProcessor.__extract_data_to_store_for_relationships(
            self.xmlfile_to_process,
            'FormDef', 'ItemGroupRef',
            'OID', 'ItemGroupOID')
        return form_to_itemgroup_data

    def extract_studyevent_to_form_data(self) -> DataFrame:
        studyevent_to_form_data: DataFrame = XmlProcessor.__extract_data_to_store_for_relationships(
            self.xmlfile_to_process,
            'StudyEventDef',
            'FormRef', 'OID', 'FormOID')
        return studyevent_to_form_data

    def extract_item_to_measurement_data(self) -> DataFrame:
        item_to_measurement_data: DataFrame = XmlProcessor.__extract_data_for_item_to_measurement_rels(
            self.xmlfile_to_process)
        return item_to_measurement_data

    def extract_item_to_rangecheck_data(self) -> DataFrame:
        item_to_rangecheck_data: DataFrame = XmlProcessor.__extract_data_for_item_to_checkvalue_rels(
            self.xmlfile_to_process)
        return item_to_rangecheck_data

    def extract_item_to_codelist_data(self) -> DataFrame:
        item_to_cl_data: DataFrame = XmlProcessor.__extract_data_to_store_for_relationships(
            self.xmlfile_to_process,
            'ItemDef',
            'CodeListRef', 'OID', 'CodeListOID')
        return item_to_cl_data

    def extract_codelist_to_codelistitem_data(self) -> DataFrame:
        cl_to_clitem_data: DataFrame = XmlProcessor.__extract_data_for_codelist_to_codelistitem_rels(
            self.xmlfile_to_process)
        return cl_to_clitem_data

    @staticmethod
    def create_fancy_dict(soup):
        data = {}
        # Iterate over all Study elements
        for study in soup.find_all('Study'):
            studyoidcandidate = study.get('OID')
            if studyoidcandidate == "some OID":
                odm = soup.find('ODM')
                study_oid = odm.get('Description')
            else:
                study_oid = study.get('OID')

            data[study_oid] = {}

            # Iterate over all ItemDef elements
            for item in study.find_all('ItemDef'):
                item_oid = item.get('OID')
                data[study_oid][item_oid] = {}

                # Iterate over all Alias elements
                for alias in item.find_all('Alias'):
                    context = alias.get('Context')
                    match = re.match(r'(.*)\[([0-9],[0-9])\]', context)
                    if match:
                        ontology_context = match.group(1)
                        xy = match.group(2)
                        if ',' in xy:
                            x, y = xy.split(',')
                            x = int(x)
                            y = int(y)
                        else:
                            x = int(xy)
                            y = None
                    else:
                        ontology_context = context
                        x = None
                        y = None

                    if ontology_context not in data[study_oid][item_oid]:
                        data[study_oid][item_oid][ontology_context] = {}
                    if x is not None:
                        if x not in data[study_oid][item_oid][ontology_context]:
                            data[study_oid][item_oid][ontology_context][x] = []
                        data[study_oid][item_oid][ontology_context][x].append(y)
        return data

    def prepare_tuple_for_postcoord_rels(self):
        mydict = XmlProcessor.create_fancy_dict(self.xmlfile_to_process)
        return mydict


def dict_to_list_of_tuples(mydictionary):
    data = [
        {
            "study_oid": study_oid,
            "item_oid": item_oid,
            "ontology": ontology,
            "x": x,
            "y": y
        }
        for study_oid, items in mydictionary.items()
        for item_oid, ontologies in items.items()
        for ontology, x_values in ontologies.items()
        for x, y_values in x_values.items()
        for y in y_values
    ]
    return data


def parse_xml(
        path_to_datafiles):  # creates XmlProcessor object for each xml file
    neo4jconnection_object = db.Neo4jConnection()
    for file in os.listdir(path_to_datafiles):
        if file.endswith(".xml"):
            filename = os.path.splitext(file)[0]
            filepath = os.path.join(path_to_datafiles, file)
            with open(filepath) as fp:
                parsed_xml = BeautifulSoup(fp, "xml")
                # create XmlProcessor object for parsed_xml files
                xmlprocessor_object = XmlProcessor(parsed_xml)
                # do stuff with the object
                study_oid = xmlprocessor_object.get_studyoid()

                # create nodes
                try:
                    study_data = xmlprocessor_object.extract_study_data()
                    neo4jconnection_object.create_study_node(study_data, filename)

                    itemgroup_data = xmlprocessor_object.extract_itemgroup_data()
                    neo4jconnection_object.create_itemgroup_node(itemgroup_data, study_oid, filename)

                    item_data = xmlprocessor_object.extract_item_data()
                    neo4jconnection_object.create_item_node(item_data, study_oid, filename)

                    studyevent_data = xmlprocessor_object.extract_studyevent_data()
                    neo4jconnection_object.create_studyevent_node(studyevent_data, study_oid, filename)

                    form_data = xmlprocessor_object.extract_form_data()
                    neo4jconnection_object.create_form_node(form_data, study_oid, filename)

                    alias_data = xmlprocessor_object.extract_item_to_alias_data()
                    neo4jconnection_object.create_alias_node(alias_data, study_oid, filename)

                    measurement_data = xmlprocessor_object.extract_measurement_data()
                    neo4jconnection_object.create_measurementunit_node(measurement_data)

                    rangecheck_data = xmlprocessor_object.extract_rangecheck_data()
                    neo4jconnection_object.create_rangecheck_node(rangecheck_data)

                    codelist_data = xmlprocessor_object.extract_codelist_data()
                    neo4jconnection_object.create_codelist_node(codelist_data, study_oid, filename)

                    codelistitem_data = xmlprocessor_object.extract_codelistitem_data()
                    neo4jconnection_object.create_clitem_node(codelistitem_data)

                    if parsed_xml.findAll('BasicDefinitions'):
                        neo4jconnection_object.create_basedef_node(study_oid, filename)

                    logger.info(f"Inserted all nodes for file {filename}.xml")

                except:
                    print(f"Nodes for file {filename}.xml could not be inserted")

                # create rels
                try:
                    itemgroup_to_item_data = xmlprocessor_object.extract_itemgroup_to_item_data
                    neo4jconnection_object.create_itemgroup_to_item_rel(itemgroup_to_item_data, study_oid, filename)

                    item_to_alias_data = xmlprocessor_object.extract_item_to_alias_data()
                    neo4jconnection_object.create_item_to_alias_rel(item_to_alias_data, study_oid, filename)

                    itemgroup_to_alias_data = xmlprocessor_object.extract_itemgroup_to_alias_data()
                    neo4jconnection_object.create_itemgroup_to_alias_rel(itemgroup_to_alias_data, study_oid, filename)

                    form_to_itemgroup_data = xmlprocessor_object.extract_form_to_itemgroup_data()
                    neo4jconnection_object.create_form_to_itemgroup_rel(form_to_itemgroup_data, study_oid, filename)

                    studyevent_to_form_data = xmlprocessor_object.extract_studyevent_to_form_data()
                    neo4jconnection_object.create_studyevent_to_form_rel(studyevent_to_form_data, study_oid, filename)

                    neo4jconnection_object.create_study_to_studyevent_rel(study_oid, filename)

                    item_to_measurement_data = xmlprocessor_object.extract_item_to_measurement_data()
                    neo4jconnection_object.create_item_to_measurement_rel(item_to_measurement_data, study_oid, filename)

                    item_to_rangecheck_data = xmlprocessor_object.extract_item_to_rangecheck_data()
                    neo4jconnection_object.create_item_to_rangecheck_rel(item_to_rangecheck_data)

                    item_to_cl_data = xmlprocessor_object.extract_item_to_codelist_data()
                    neo4jconnection_object.create_item_to_codelist_rel(item_to_cl_data, study_oid, filename)

                    cl_to_clitem_data = xmlprocessor_object.extract_codelist_to_codelistitem_data()
                    neo4jconnection_object.create_codelist_to_codelistitem_rel(cl_to_clitem_data, study_oid, filename)

                    # rels to and from BasicDefinitions node
                    neo4jconnection_object.create_basedef_to_study_rel(study_oid, filename)
                    neo4jconnection_object.create_basedef_to_measurement_rel(measurement_data, study_oid, filename)

                except:
                    print(f"Relationships for file {filename}.xml could not be inserted")

                logger.info(f"Inserted all rels for file {filename}.xml")

                # postprocess
                neo4jconnection_object.postprocess_set_label_post()
                dictionary = xmlprocessor_object.prepare_tuple_for_postcoord_rels()
                datatuple = dict_to_list_of_tuples(dictionary)
                neo4jconnection_object.postprocess_connect_postcoord_aliases(datatuple)

                logger.info(f"Post-processing done for file {filename}.xml")
